/*
  medx_anovanc - fMRI ANOVA analysis of non-cyclic data

  Stuart Clare, FMRIB Physics Group

  Copyright (c) 1998-2000 University of Oxford
   
  Clare, S. et.al. (1999) Magnetic Resonance in Medicine. 42,1117-1122.
*/

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "hfunc.h"
#include "ftoz.h"

/* usage: medx_anovanc infile Foutfile Zoutfile %threshold events eventsize eventfile [-av filename]*/

#define MAXLEN 128

int main(int argc,char *argv[])
{
  char infile[MAXLEN],foutfile[MAXLEN],zoutfile[MAXLEN];
  char trigfile[MAXLEN],avfile[MAXLEN];
  short *image,thold;
  short max=0;
  double *F,*Z,thresh;
  int i,j,k,l,points,cycles,size,av=0;
  int ppl,lpi,ipv,vols,type,*trigs;
  long volsize,c;
  float vx,vy,vz,*fdata,*average=NULL;
  double *data,T,C,SST,SSTr,Ti,SSE;
  FILE *ifp,*fofp,*zofp,*fp,*afp=NULL;

  if(argc<8){
    printf("usage: medx_anovanc infile Ffile Zfile %%thold events eventsize eventfile [-av file]\n");
    return(1);
  }

  sprintf(infile,"%s.img",argv[1]);
  sprintf(foutfile,"%s.img",argv[2]);
  sprintf(zoutfile,"%s.img",argv[3]);
  sscanf(argv[4],"%lf",&thresh);
  sscanf(argv[5],"%d",&cycles);
  sscanf(argv[6],"%d",&points);
  sscanf(argv[7],"%s",trigfile);
  if(argc>8){
    if(!strcmp(argv[8],"-av")){
      av=1;
      sprintf(avfile,"%s.img",argv[9]);
    }
    else av=0;
  }
  else av=0;
  /* Read header */
  if(hread_plus(infile,&ppl,&lpi,&ipv,&vols,&type,&vx,&vy,&vz)==-1){
    printf("Cannot read header file: %s\b\b\b\b.hdr",infile);
    return(1);
  }
  volsize=ppl*lpi*ipv;

  if((trigs=(int *)malloc(cycles*sizeof(int)))==NULL){
    printf("Malloc failed.");
    return(1);
  }
  size=points*cycles;
  if(size>vols){
    printf("Events times event size greater than number of volumes");
    return(1);
  }

  if((fp=fopen(trigfile,"r"))==NULL){
    printf("Can't open: %s",trigfile);
    return(1);
  }
  for(i=0;i<cycles;i++){
    fscanf(fp,"%d\n",&trigs[i]);
    if((trigs[i]+points)>vols){
      printf("Not enough images for event file");
      return(1);
    }
  }
  /* Open files and malloc */
  if((ifp=fopen(infile,"r"))==NULL){
    printf("Can't open : %s\n",infile);
    return(1);
  }
  if((fofp=fopen(foutfile,"wb"))==NULL){
    printf("Can't open file: %s\n",foutfile);
    return(1);
  }
  if((zofp=fopen(zoutfile,"wb"))==NULL){
    printf("Can't open file: %s\n",zoutfile);
    return(1);
  }
  if(av){
    if((afp=fopen(avfile,"wb"))==NULL){
      printf("Can't open file: %s\n",avfile);
      return(1);
    }
  }
  if(volsize>ppl*vols){
    if((image=(short *)malloc(volsize*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  } else {
    if((image=(short *)malloc(ppl*vols*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }
  if((data=(double *)malloc(vols*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((F=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((Z=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((fdata=(float *)malloc(volsize*sizeof(float)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if(av){
    if((average=(float *)malloc(points*volsize*sizeof(float)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }

  /* Read in first image and find max */
  fseek(ifp,0,SEEK_SET);
  if((fread(image,sizeof(short),volsize,ifp))!=volsize){
    printf("Read error.\n");
    return(1);
  }
  for(i=0;i<volsize;i++)if(image[i]>max)max=image[i];
  thold=max*thresh/100;

  /* For each line in image */
  for(i=0;i<ipv*lpi;i++){	
    /* Read in data */	
    for(k=0;k<vols;k++){
      fseek(ifp,((k*volsize)+(i*ppl))*sizeof(short),0);
      if((fread(&image[k*ppl],sizeof(short),ppl,ifp))!=ppl){
	printf("Read error.\n");
	return(1);
      }
    }
		
    /* For each point in line */
    for(j=0;j<ppl;j++){
      if(image[j]>thold){
	for(k=0;k<cycles;k++){
	  for(l=0;l<points;l++){
	    data[(k*points)+l]=(double)image[((trigs[k]+l)*ppl)+j];
	  }
	}

	T=SST=0.0;
	for(k=0;k<size;k++){
	  T+=data[k];
	  SST+=(data[k]*data[k]);
	}
	C=(T*T)/size;
	SST-=C;
	SSTr=0.0;
	for(k=0;k<points;k++){
	  for(l=0,Ti=0.0;l<cycles;l++){
	    Ti+=data[(l*points)+k];
	  }
	  SSTr+=(Ti*Ti);
	}
	SSTr/=cycles;
	SSTr-=C;
	SSE=SST-SSTr;
	F[(i*ppl)+j]=(SSTr/(points-1))/(SSE/(points*(cycles-1)));
      }
      else F[(i*ppl)+j]=0;
    }
  }
  /* Create average image */
  if(av){
    for(c=0;c<volsize*points;c++)average[c]=0.0;
    for(j=0;j<cycles;j++){
      fseek(ifp,trigs[j]*volsize*sizeof(short),SEEK_SET);
      for(k=0;k<points;k++){
	if((fread(image,sizeof(short),volsize,ifp))!=volsize){
	  printf("Read error.\n");
	  return(1);
	}
	for(c=0;c<volsize;c++)average[k*volsize+c]+=((float)image[c]/(float)cycles);
      }
    }
    if(fwrite(average,sizeof(float),volsize*points,afp)!=volsize*points){
      printf("Write error\n");
      return(1);
    }
    if(hmake_plus(avfile,ppl,lpi,ipv,points,DT_FLOAT,(float)max,0,vx,vy,vz)==-1){
      printf("Cannot create header: %s\b\b\b\b.hdr\n",foutfile);
      return(1);
    }
  }

  /* Transform F to Z */
  ftoz(F,Z,volsize,points-1,points*(cycles-1));

  /* Write statistics and create header */		
  for(c=0;c<volsize;c++)fdata[c]=(float)F[c];
  fseek(fofp,0,0);
  if(fwrite(fdata,sizeof(float),volsize,fofp)!=volsize){
    printf("Write error\n");
    return(1);
  }	
  if(hmake_plus(foutfile,ppl,lpi,ipv,1,DT_FLOAT,10,0,vx,vy,vz)==-1){
    printf("Cannot create header: %s\b\b\b\b.hdr\n",foutfile);
    return(1);
  }

  for(c=0;c<volsize;c++)fdata[c]=(float)Z[c];
  fseek(zofp,0,0);
  if(fwrite(fdata,sizeof(float),volsize,zofp)!=volsize){
    printf("Write error\n");
    return(1);
  }	
  if(hmake_plus(zoutfile,ppl,lpi,ipv,1,DT_FLOAT,10,0,vx,vy,vz)==-1){
    printf("Cannot create header: %s\b\b\b\b.hdr\n",zoutfile);
    return(1);
  }
  return(0);
}
