/* {{{ Copyright etc. */

/*  avwstats.c - basic image stats measurements

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss/libss.h"
#include "libss/libavw.h"

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("Usage: avwstats <input> [options]\n");
  printf("-r : <robust min intensity> <robust max intensity>\n");
  exit(1);
}

/* }}} */
/* {{{ main */

int main(argc, argv)
  int   argc;
  char  *argv [];
{
  /* {{{ vars */

image_struct im;
int x_size, y_size, z_size, i;
float scale;

/* }}} */

  /* {{{ process arguments */

if (argc<3)
     usage();

avw_read(argv[1],&im);

x_size=im.x; y_size=im.y; z_size=im.z;
scale=MIN(im.xv,MIN(im.yv,im.zv)); /* minimum voxel side length */

for (i = 2; i < argc; i++) {
  if (!strcmp(argv[i], "-r"))
    /* {{{ range */

{
  im.min=im.max=0;
  find_thresholds(&im,0.1);
  printf("%f %f ",(double)im.thresh2,(double)im.thresh98);

  /* {{{ COMMENT  */

#ifdef FoldingComment

c_of_g (im,&cgx,&cgy,&cgz);
cgx*=im.xv; cgy*=im.yv; cgz*=im.zv;
printf("CofG (%f,%f,%f) mm\n",cgx,cgy,cgz);

radius = find_radius (im,im.xv*im.yv*im.zv);
printf("RADIUS %f\n",radius);

{
  FDT *tmpimage = (FDT *) malloc(sizeof(FDT)*x_size*y_size*z_size);

  i=0;
  for(z=0; z<z_size; z++)
    for(y=0; y<y_size; y++)
      for(x=0; x<x_size; x++)
	{
	  FDT tmp=IA(in,x,y,z);
	  
	  if ( (tmp>thresh2) && (tmp<thresh98) &&
	       ( (x*im.xv-cgx)*(x*im.xv-cgx) + (y*im.yv-cgy)*(y*im.yv-cgy) + (z*im.zv-cgz)*(z*im.zv-cgz) < radius*radius ) )
	    tmpimage[i++]=tmp;
	}
  medianval = median(0.5,tmpimage,i);
  printf("MEDIANVAL %f\n",(double)medianval);
  free(tmpimage);
}

#endif

/* }}} */
}

/* }}} */
  else
    usage();
}

/* {{{ COMMENT  */

#ifdef FoldingComment
for (i = 3; i < argc; i++) {
  if (!strcmp(argv[i], "-x"))
    output_xtopol=1;
  else if (!strcmp(argv[i], "-c"))
    output_cost=1;
  else if (!strcmp(argv[i], "-s"))
    output_skull=1;
  else if (!strcmp(argv[i], "-S"))
    /* {{{ colour-coded skull image */

    {
      output_skull=1;
      code_skull=1;
    }

/* }}} */
  else if (!strcmp(argv[i], "-m"))
    output_mask=1;
  else if (!strcmp(argv[i], "-o"))
    output_overlay=1;
  else if (!strcmp(argv[i], "-n"))
    output_brain=0;
  else if (!strcmp(argv[i], "-t"))
    apply_thresholding=1;
  else if (!strcmp(argv[i], "-f"))
    /* {{{ fractional brain threshold */

    {
      i++;
      if (argc<i+1) /* option following f hasn't been given */
      {
	printf("Error: no value given following -f\n");
	usage();
      }
      brain_threshold=atof(argv[i]);
      if ( (brain_threshold<=0) || (brain_threshold>=1) )
      {
	printf("Error: value following -f must lie between 0 and 1\n");
	usage();
      }
    }

/* }}} */
  else if (!strcmp(argv[i], "-g"))
    /* {{{ gradient fractional brain threshold */

    {
      i++;
      if (argc<i+1) /* option following g hasn't been given */
      {
	printf("Error: no value given following -g\n");
	usage();
      }
      gradthresh=atof(argv[i]);
      if ( (gradthresh<-1) || (gradthresh>1) )
      {
	printf("Error: value following -g must lie between -1 and 1\n");
	usage();
      }
    }

/* }}} */
  else
    usage();
}

#endif

/* }}} */

/* }}} */

  printf("\n");
  return(0);
}

/* }}} */
