/*  mjimage.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Quick and dirty image storage class


#include "mjimage.h"
#include "miscmaths.h"
#include <iostream>
#include <vector>
#include <algorithm>

#ifndef NO_NAMESPACE
 using namespace MISCMATHS;

 namespace MJIMAGE {
#endif

  basicvolume::basicvolume() : Data(0), data_owner(false)
    {
      this->initialize(0,0,0,0);
    }


  basicvolume::basicvolume(const basicvolume& source) : Data(0), data_owner(false)
    {
      this->initialize(source.columns(),source.rows(),source.slices(),0);
      this->copy(source);
    }

  basicvolume::basicvolume(int column, int row, int slice, float *d) : Data(d), data_owner(false)
    {
      this->initialize(column,row,slice,d);
    }

  int basicvolume::initialize(int column, int row, int slice, float *d)
    {
      if (data_owner && (Data!=0))  delete [] Data;
      set_whole_cache_validity(false);
      Xsize = 1.0;
      Ysize = 1.0;
      Zsize = 1.0;
      Xsign = 1;
      Ysign = 1;
      Zsign = 1;
      SlicesZ = slice;
      RowsY = row;
      ColumnsX = column;
      Bound = SlicesZ * RowsY * ColumnsX;
      SliceOffset = RowsY * ColumnsX;
      if (Bound > 0) {
	if (d != 0) {
	  Data = d;
	  data_owner = false;
	}
	else {
	  Data = new float[Bound];
	  data_owner = true;
	}
      } else {
	Data = 0;
      }
      return 0;
    }


  int basicvolume::reinitialize(int column, int row, int slice, float *d)
    {
      return this->initialize(column,row,slice,d);
    }


  int basicvolume::reinitialize(const basicvolume& source)
    {
      if (source.Data==0) return -1;
      this->initialize(source.columns(),source.rows(),source.slices(),0);
      return this->copy(source);
    }



  basicvolume::~basicvolume()
    {  if (data_owner) delete [] Data;}


  float basicvolume::set(float val)
    {
      set_whole_cache_validity(false);
      for (unsigned int i=0; i<Bound; i++) {
	Data[i] = val;
      }
      return val;
    }


  float basicvolume::operator=(float val)
  {
    this->set(val);
    return val;
  }


  int basicvolume::copy(const basicvolume& source)
    {
      if (source.Data==0)  return -1;
      Xsize = source.Xsize;
      Ysize = source.Ysize;
      Zsize = source.Zsize;
      Xsign = source.Xsign;
      Ysign = source.Ysign;
      Zsign = source.Zsign;
      int NS=Min(SlicesZ,source.slices());
      int NR=Min(RowsY,source.rows());
      int NC=Min(ColumnsX,source.columns());
      for (int z=0; z<NS; z++) {
	for (int y=0; y<NR; y++) {
	  for (int x=0; x<NC; x++) {
	    this->operator()(x,y,z) = source(x,y,z);
	  }
	}
      }
      this->copylazymanager(source);
      set_whole_cache_validity(false);
      return 0;
    }
  

  ///////////////////////////////////////////////////////////////////////////

  //----------------------------------------------------------------------//

  int volume_get_cog(const volume& v, ColumnVector& cog)
    {
      cog(1)=0.0;
      cog(2)=0.0;
      cog(3)=0.0;
      float val=0.0, total=0.0, vmin=v(0,0,0);
      unsigned int zb = v.slices(), yb = v.rows(), xb = v.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    val = v(x,y,z);
	    cog(1) += val*x;
	    cog(2) += val*y;
	    cog(3) += val*z;
	    total += val;
	    if (val < vmin)   vmin = val;
	  }
	}
      }
      // correct for possible negative intensities
      total -= vmin * xb*yb*zb;
      cog(1) -= vmin * (xb-1)*xb*yb*zb/2;
      cog(2) -= vmin * (yb-1)*xb*yb*zb/2;
      cog(3) -= vmin * (zb-1)*xb*yb*zb/2;
      if (fabs(total) < 1e-5) {
	cerr << "WARNING::in calculating COG, total = 0.0" << endl;
	total = 1.0;
      }
      cog(1) /= total;
      cog(2) /= total;
      cog(3) /= total;
      // Now convert these values from voxel to physical space
      ColumnVector cog4(4);
      cog4.Rows(1,3) = cog.Rows(1,3);
      cog4(4) = 1.0;
      cog4 = v.sampling_matrix() * cog4;
      cog.Rows(1,3) = cog4.Rows(1,3);
      return 0;
    }


  ColumnVector calc_cog(const volume& vol)
    {
      ColumnVector v_cog(3);
      volume_get_cog(vol,v_cog);
      return v_cog;
    }



  //-------------------------------------------------------------------//

  
  float calc_bval_general(const volume& vol, unsigned int edgewidth)
    {
      unsigned int zb = vol.slices(), yb = vol.rows(), xb = vol.columns();
      unsigned int ew = edgewidth, numbins;
      numbins = 2*(xb-2*ew)*(yb-2*ew)*ew + 2*(xb-2*ew)*zb*ew + 2*yb*zb*ew;
      std::vector<float> hist(numbins);
      // put the edge voxel values into the histogram
      unsigned int hindx = 0;
      for (unsigned int e=0; e<ew; e++) {
	// put in the faces
	// xy faces (small lids of the box)
	for (unsigned int x=ew; x<xb-ew; x++) {
	  for (unsigned int y=ew; y<yb-ew; y++) {
	  hist[hindx++] = vol.operator()(x,y,e);
	  hist[hindx++] = vol.operator()(x,y,zb-1-e);
	  }
	}
	// xz faces (smallish edge faces)
	for (unsigned int x=ew; x<xb-ew; x++) {
	  for (unsigned int z=0; z<zb; z++) {
	  hist[hindx++] = vol.operator()(x,e,z);
	  hist[hindx++] = vol.operator()(x,yb-1-e,z);
	  }
	}
	// yz faces (large edge faces)
	for (unsigned int y=0; y<yb; y++) {
	  for (unsigned int z=0; z<zb; z++) {
	  hist[hindx++] = vol.operator()(e,y,z);
	  hist[hindx++] = vol.operator()(xb-1-e,y,z);
	  }
	}
      }
      sort(hist.begin(),hist.end());
      unsigned int percentile10 = numbins / 10;
      float v10 = hist[percentile10];
      return v10;
    }

  float calc_bval(const volume& vol) {
    return calc_bval_general(vol,2);
  }
 


  //---------------------------------------------------------------------//
   


  volume::volume() 
    {
      this->volume::initialize();
    }
  
  volume::volume(int column, int row, int slice, float *d) :
    basicvolume(column,row,slice,d)
    {
      this->volume::initialize();
    }

  volume::volume(const volume& source)
    {
      this->volume::initialize();
      this->volume::reinitialize(source);
    }


  volume::~volume() 
    { 
      // do nothing
    }

  int volume::initialize()
    {
      v_sampling_matrix = Identity(4);
      cog.init(this,calc_cog);
      backgroundval.init(this,calc_bval);
      avw_origin.ReSize(3);
      avw_origin = 0;
      return 0;
    }

  volume& volume::operator=(const volume& source)
  { 
    this->volume::reinitialize(source);
    return *this;
  }
  
  int volume::reinitialize(int column, int row, int slice, float *d)
    {
      this->basicvolume::reinitialize(column,row,slice,d);
      this->volume::initialize();
      set_whole_cache_validity(false);
      return 0;
    }

  int volume::reinitialize(const volume& source) 
    {
      this->basicvolume::reinitialize(source);
      v_sampling_matrix = source.v_sampling_matrix;
      // copy lazy properties
      cog.copy(source.cog,this);
      backgroundval.copy(source.backgroundval,this);
      avw_origin = source.avw_origin;
      //this->setvoxelsize(this->getx(),this->gety(),this->getz()); // ??
      return 0;
    };
  
  Matrix volume::sampling_matrix() const
    {
      return v_sampling_matrix;
    }

  void volume::set_sampling_matrix(const Matrix& samplingmat) 
    {
      v_sampling_matrix = samplingmat;
    }


  //-----------------------------------------------------------------------//

  void volume::setvoxelsize(const float x, const float y, const float z)
    { 
      // set the diagonal components of the v_sampling_matrix to the voxelsize
      // AND make sure that if A0*X = 0 before, then A1*X = 0 after
      //  (that is, V = A0^-1 * [0 0 0 1]' is voxelorigin and
      //     for A1(3,3) * V = T1  =>  -T1 is the appropriate translation)
      ColumnVector t1(3), voxelorigin(4), worldorigin(4);
      worldorigin << 0.0 << 0.0 << 0.0 << 1.0;
      voxelorigin = v_sampling_matrix.i() * worldorigin;
      this->setx(x);  this->sety(y);  this->setz(z);
      v_sampling_matrix(1,1) = this->getx();
      v_sampling_matrix(2,2) = this->gety();
      v_sampling_matrix(3,3) = this->getz();
      t1 = v_sampling_matrix.SubMatrix(1,3,1,3) * 
	             voxelorigin.SubMatrix(1,3,1,1);
      v_sampling_matrix.SubMatrix(1,3,4,4) = -t1;
    }
  
  void volume::setvoxelorigin(const float x, const float y, const float z)
    {
      ColumnVector trans0(3), trans1(3);
      trans0 << x << y << z;
      trans1 = v_sampling_matrix.SubMatrix(1,3,1,3) * trans0;
      v_sampling_matrix.SubMatrix(1,3,4,4) = -trans1;
    }

  ColumnVector volume::getvoxelorigin() const
    {
      ColumnVector oret(3), voxelorigin(4), worldorigin(4);
      worldorigin << 0.0 << 0.0 << 0.0 << 1.0;
      voxelorigin = v_sampling_matrix.i() * worldorigin;
      oret = voxelorigin.SubMatrix(1,3,1,1);
      return oret;
    }

  //=====================================================================//

  tvolume::tvolume()
    {
      this->initialize(NULL, 0, 0);
    }
  
  tvolume::~tvolume() {}

  int tvolume::initialize(const volume* source, const int len, const float TR)
    {
      tr = TR;
      length = len;
      
      if (source == NULL){
	vols = NULL;
	return 0;
      }
      
      vols = new volume[len];
      
      for (int i=0; i < len; i++)
	vols[i] = source[i];
      
      return 0;
    }

  int tvolume::reinitialize(const int len, const float TR)
    {
      tr = TR;
      length = len;

      vols = new volume[len];
      
      return 0;
    }
      

  int tvolume::reinitialize(const volume* source, const int len, const float TR)
    {
      return this->initialize(source, len, TR);
    }

  
#ifndef NO_NAMESPACE
 }
#endif










