/*  mjimage.h

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Quick and dirty image storage class


#if !defined(__mjimage_h)
#define __mjimage_h

#include <cstdlib>
#include "newmatap.h"
#include "lazy.h"

#ifndef NO_NAMESPACE
 using namespace NEWMAT;
 using namespace LAZY;

 namespace MJIMAGE {
#endif

  class basicvolume : public lazymanager {
  private:
    int SlicesZ;
    int RowsY;
    int ColumnsX;
    float Xsize;
    float Ysize;
    float Zsize;
    int Xsign;
    int Ysign;
    int Zsign;
    unsigned int Bound;
    int SliceOffset;
    float *Data;
    bool data_owner;

    int initialize(int column, int row, int slice, float *d=0);
    float set(float val);
  public:
    basicvolume();
    basicvolume(const basicvolume& source);
    basicvolume(int column, int row, int slice, float *d=0);
    ~basicvolume();

    int reinitialize(int column, int row, int slice, float *d=0);
    int reinitialize(const basicvolume& source);  // copy WITH resizing

    inline const int slices(void) const { return SlicesZ; }
    inline const int rows(void) const { return RowsY; }
    inline const int columns(void) const { return ColumnsX; }
    inline const int xsize(void) const { return ColumnsX; }
    inline const int ysize(void) const { return RowsY; }
    inline const int zsize(void) const { return SlicesZ; }
    inline const float getx(void) const { return Xsize; }
    inline const float gety(void) const { return Ysize; }
    inline const float getz(void) const { return Zsize; }
    inline const float xsign(void) const { return Xsign; }
    inline const float ysign(void) const { return Ysign; }
    inline const float zsign(void) const { return Zsign; }
    float operator=(float val);
    basicvolume& operator=(const basicvolume& source)   // copy WITH resizing
                        { this->reinitialize(source); return *this; }

    
     inline float& operator()(int x, int y, int z)
      { 
	set_whole_cache_validity(false); 
	return *(Data + (z*RowsY + y)*ColumnsX + x); 
      }
     inline float& operator()(unsigned int x, unsigned int y, unsigned int z)
      { 
	set_whole_cache_validity(false); 
	return *(Data + (z*RowsY + y)*ColumnsX + x); 
      }
     inline  float& operator()(int x, int y, int z)
      const { return *(Data + (z*RowsY + y)*ColumnsX + x); }
     inline  float& operator()(unsigned int x, unsigned int y, unsigned int z)
      const { return *(Data + (z*RowsY + y)*ColumnsX + x); }

     inline void getneighbours(const int x, const int y, const int z, 
			       float &v000, float &v001, float &v010,
			       float &v011, float &v100, float &v101,
			       float &v110, float &v111) const {
       float *ptr = Data + (z*RowsY + y)*ColumnsX + x;
       v000 = *ptr;
       ptr++; 
       v100 = *ptr;
       ptr+= ColumnsX;  
       v110 = *ptr;
       ptr--; 
       v010 = *ptr;
       ptr += SliceOffset;
       v011 = *ptr;
       ptr++;
       v111 = *ptr;
       ptr-= ColumnsX;
       v101 = *ptr;
       ptr--;
       v001 = *ptr;
     }
		     

    inline bool in_bounds(int x, int y, int z) const
      { return ( (x>=0) && (y>=0) && (z>=0) && (x<ColumnsX)
		 && (y<RowsY) && (z<SlicesZ) ); }
    inline bool in_bounds(unsigned int x, unsigned int y, unsigned int z) const
      { return ( (x>=0) && (y>=0) && (z>=0) && (x<(unsigned int) ColumnsX)
		 && (y<(unsigned int) RowsY) && (z<(unsigned int) SlicesZ) ); }

  protected:
    int copy(const basicvolume& source);  // copy WITHOUT resizing
    void setx(const float x) { Xsize = fabs(x);  if (x<0) Xsign=-1; }
    void sety(const float y) { Ysize = fabs(y);  if (y<0) Ysign=-1; }
    void setz(const float z) { Zsize = fabs(z);  if (z<0) Zsign=-1; }
  };

  ///////////////////////////////////////////////////////////////////////////

  
  class volume : public basicvolume {
  private:
    Matrix v_sampling_matrix;

    int initialize();
    void set_sampling_matrix(const Matrix& samplingmat);

  public:
    lazy<ColumnVector, volume> cog;
    lazy<float, volume> backgroundval;

    mutable ColumnVector avw_origin;

    volume();
    volume(const volume& source);
    volume(int column, int row, int slice, float *d=0);
    ~volume();

    float operator=(float val)  { return basicvolume::operator=(val); };
    volume& operator=(const volume& source);

    int reinitialize(int column, int row, int slice, float *d=0);
    int reinitialize(const volume& source);  // copy WITH resizing

    Matrix sampling_matrix() const;

    void setvoxelsize(const float x, const float y, const float z); 
    void setvoxelorigin(const float x, const float y, const float z);
    void setvoxelorigin(const ColumnVector& orig)
      {this->setvoxelorigin((float) orig(1),(float) orig(2),(float) orig(3));};
    ColumnVector getvoxelorigin() const;

  };



  ///////////////////////////////////////////////////////////////////////////


  class tvolume {
  private:
    volume* vols;
    float tr;
    int length;

    int initialize(const volume* source, const int len, const float TR);

  public: 
    tvolume();
    ~tvolume();

    int reinitialize(const volume* source, const int len, const float TR);
    int reinitialize(const int len, const float TR);

    inline volume& operator[](int index) const { return vols[index]; } 
    inline const float getTR(void) const { return tr; }
    inline const int getSize(void) const { return length; }
  };


#ifndef NO_NAMESPACE
 }
#endif

#endif

